package Sophos::ValidatedFetch;

use strict;
use warnings;

use URI;
use Sophos::ValidatedFetch::FileCache;
use Sophos::ValidatedFetch::Manifest;
use Sophos::ValidatedFetch::Downloader;
use PPM::PureMessage::Utils qw(pmx_user);

our $VERSION = "5.601";
our $DEBUG_VALIDATED_FETCH = 1;
our @TRACE = ();

sub DEBUG {
   push(@TRACE,join('',@_)) if $ENV{DEBUG_VALIDATED_FETCH}; 
   #warn '>>>> ' . join('',@_) . "\n" if $DEBUG_VALIDATED_FETCH;
}

sub trace {
    my @TEMP = @TRACE;
    @TRACE = ();
    return @TEMP;
}

sub new {
    my ($class, %args) = @_;
    my $self =  bless {}, $class;

    $self->{repo} = delete($args{repo});
    $self->{path} = delete($args{path});
    $self->{trace} = [];

    if ($self->{path}) {
        $self->{path_downloader}    = "$self->{path}/fetched_files";
        $self->{path_manifest}      = "$self->{path}/sophos_manifest_cache";
        $self->{path_file_cache}    = "$self->{path}/sophos_fetched_file_cache";
    } else {

        die "You must pass in path_downloader, path_manifest, and path_file_cache"
             unless scalar(grep $_, @args{qw(path_downloader path_manifest path_file_cache)}) == 3;

        $self->{path_downloader}    = delete($args{path_downloader});
        $self->{path_manifest}      = delete($args{path_manifest});
        $self->{path_file_cache}    = delete($args{path_file_cache});
        
    }

    $self->{test_no_validate}   = delete($args{test_no_validate});

    if ($self->{test_no_validate}) {
        warn <<EOF;
***********************************************************************
*** TEST OBJECT INSANTIATED WITHOUT CERTIFICATE VALIDATION          ***
***********************************************************************

EOF
    }
    
    my %dl_extra;
    my $ua = delete($args{ua});
    $dl_extra{ua} = $ua if $ua;

    $self->{downloader} = Sophos::ValidatedFetch::Downloader->new(
            repo => $self->{repo},
            path => "$self->{path_downloader}",
            %dl_extra
            );

    # Change ownership of manifest files if UID is 0
    my %manifest_file_opts = ();
    if ($> == 0) {
        my ($uid,$gid) = (pmx_user())[2,3];
        %manifest_file_opts = (
             uid => $uid,
             gid => $gid,
        );
    }

    $self->{manifests} = {};
    $self->{error} = [];
    $self->{manifest_cache} = Sophos::ValidatedFetch::FileCache->new(
            path => "$self->{path_manifest}",
            %manifest_file_opts,
            );
    $self->{file_cache} = Sophos::ValidatedFetch::FileCache->new(
            path => "$self->{path_file_cache}",
            );
    $self->{top_level} = $self->_manifest('manifest.dat');
    return $self;

}

sub delay {
   my $self = shift;
   DEBUG("waiting for 1 second");
   sleep 1;
}

sub _manifest {
    my ($self,$name) = @_;
    return Sophos::ValidatedFetch::Manifest->new( filename   => $name, 
                                                  downloader => $self->{downloader}, 
                                                  filecache  => $self->{manifest_cache},
                                                  test_no_validate => $self->{test_no_validate},
                                                 );
}

sub _addError {
    my ($self,$error) = @_;
    push(@{$self->{error}},$error);
}

sub error {
    my ($self, %args) = @_;
    my @errors = @{$self->{error}};
    $self->{error} = [];
    return @errors if wantarray;
    return join("\n",@errors);
}

sub verify_file {
    my ($self, $uri) = @_;

    my $URI_OBJECT = URI->new($uri);
    my @paths = $URI_OBJECT->path_segments();
    my $short_filename = $paths[-1];

    if(-e '/etc/puremessage-novalidation' || $ENV{PMX_NO_VALIDATE_PACKAGES}) {
        warn "Package validation is disabled.\n";
        # XXX TODO this could fail...
        return $self->{downloader}->getURI($uri);
    }

    my $top_level = $self->{top_level};
    
    unless( $top_level->validSigning() ) {
        $top_level->clear();
        DEBUG("Clearing top level manifest due to invalid signing");
    }
    # TODO XXX validate the sha1 of submanifests before using them!!!!

    my %contents = $top_level->contents();

    foreach my $manifest_name (reverse sort keys %contents) {
        
        $self->{manifests}->{$manifest_name} = $self->_manifest($manifest_name);
        my $sha1 = $self->{manifests}->{$manifest_name}->sha1;
        my $top_sha1 = $top_level->getContainedSHA1($manifest_name);

        unless($sha1 and $top_sha1 and ($top_sha1 eq $sha1)) {
          $self->{manifests}->{$manifest_name}->clear();
          # the cached manifest was bad, discard it
        }
    }

    my $downloader     = $self->{downloader};
    my $cache          = $self->{file_cache};
    my $RETRY_ATTEMPTS = 3;
    my $VALID          = 0;
    my $file_path_of_download = '';

    RETRY_MANIFEST_LOOP: while ( ($RETRY_ATTEMPTS--) and !$VALID ) {

        # repeat until file is downloaded and verified correctly

        $file_path_of_download = $downloader->getURI($uri);
        my $download_error = $downloader->error();

        if( $download_error or (not $file_path_of_download) ) {

          DEBUG("download of $uri failed: $download_error");
          $self->_addError( $download_error );
          $self->delay();

          next RETRY_MANIFEST_LOOP;

        }

        DEBUG("download of $uri to $file_path_of_download");
        
        $cache->store($file_path_of_download);
        my $downloaded_file_sha1 = $cache->sha1($short_filename);

        DEBUG("file $file_path_of_download has a sha1 of " . $cache->sha1($short_filename) );
        DEBUG("There are no sub manifests") unless scalar(keys %contents);
        
        # emulate "test if valid", refresh, "test if valid" using a foreach 
        foreach my $is_first_loop ( 0..1 ) {
            foreach my $manifest ( reverse sort keys %{$self->{manifests}} ) {
                # see if we can find the file in each of the sub manifests, if we find it check if it has the right sha1

                my $expected_sha1 = $self->{manifests}->{$manifest}->getContainedSHA1($short_filename);

                 if ( $expected_sha1 and $downloaded_file_sha1 and ($expected_sha1 eq $downloaded_file_sha1) ) {

                     DEBUG("file $file_path_of_download has the correct sha1 from manifest '$manifest', file is VALID");
                     $VALID = 1;
                     last RETRY_MANIFEST_LOOP;

                 } elsif( not defined($expected_sha1)) {

                     DEBUG("file $file_path_of_download is not contained in manifest '$manifest'");
                     next;

                 } else {
                     DEBUG("file $file_path_of_download has the wrong sha1 of $downloaded_file_sha1 but manifest '$manifest' wanted $expected_sha1");
                 }

            }

            $self->refresh_manifests() unless $is_first_loop; # couldn't find it or it had the wrong sha1, refresh and try again
        }

        $cache->remove($short_filename); # delete our cached copy of that file because it may be the reason we couldn't validate

     } # try the above 3 times or while not valid

     if ($VALID) {

         my $return_path = $cache->file_path($short_filename);
         DEBUG("verify_file has was successful returning the filepath '$return_path'\n");
         return $return_path;

     } else {

         DEBUG("verify_file has failed for file $uri");
         return;
     }
}

sub refresh_manifests {
    my $self = shift;

    my %found   = (); # the new list hash of manifests found within the toplevel
    my $failure = 1;
    my $RETRY   = 0;
    my $error;
RETRY_LOOP: while ( $failure and ($RETRY < 3) ) {

        %found   = ();
        $failure = 0;

        DEBUG("refreshing top level manifest");
        $self->{top_level}->clear();
        $self->{top_level}->refresh();
        $error = $self->{top_level}->error();

        if ( $error ) {

           DEBUG("download of top level manifest had errors: $error");
           $self->_addError($error);
           $failure++;
           $RETRY++;
           $self->delay(); 
           next RETRY_LOOP;
        }

        if( not $self->{top_level}->validSigning  ) {

            DEBUG("top level manifest did not have a valid certificate");
            $error = $self->{top_level}->error();
            DEBUG("top level error is: $error");
            $self->_addError($error);
            $failure++;
            $RETRY++;
            # we don't wait on invalid cert errors
            next RETRY_LOOP;
        } else {
            DEBUG("Signing all good!");
        }

        my %contents = $self->{top_level}->contents();

        foreach my $manifest ( reverse sort keys %contents) {

          DEBUG("inspecting '$manifest' to see if it needs refreshing");
          my $sub_manifest;

          if ( not exists($self->{manifests}->{$manifest}) ) {
              DEBUG("manifest '$manifest' had to be created, done.");
              $sub_manifest = $self->_manifest($manifest); # create it
          } else {
              $sub_manifest = $self->{manifests}->{$manifest};
          }
          
          my $sha1 = $sub_manifest->sha1;

          if ( (not defined($sha1)) or ($sha1 ne $contents{$manifest}->{sha1}) ) {

              $sha1 = '' unless $sha1; # just for error messages
              DEBUG("top level manifest says sha1 of $manifest should be '$contents{$manifest}->{sha1}' but it actually is: '$sha1'");

              $sub_manifest->refresh();
              $error = $sub_manifest->error();
              DEBUG("cleared and refreshed '$manifest'");

              my $sub_manifest_sha1 = $sub_manifest->sha1;

              if ( $error ) {

                  DEBUG("there was an error downloading manifest '$manifest': $error");

                  $failure = 1;
                  $self->_addError($error);
                  $sub_manifest->clear();
                  $self->delay();

              } elsif ( (not defined($sub_manifest_sha1))  or  ($sub_manifest_sha1 ne $contents{$manifest}->{sha1}) ) {
                  
                  $sub_manifest_sha1 = '' unless $sub_manifest_sha1; # for debug statement
                  DEBUG("and the top level sha1 still doesn't match $manifest sha1 of '$sub_manifest_sha1': $sub_manifest->{data}");
                  $failure = 1;
                  $self->delay();
                  $sub_manifest->clear();

              } else {
                  DEBUG("manifest '$manifest' now matches the top level sha1");
              }

          } else {

              $sha1 = $sub_manifest->sha1;
              DEBUG("manifest '$manifest' has a sha1 '$sha1' that matches our top level manifest '$contents{$manifest}->{sha1}");

          }

          # build up a new list of sub manifests
          $found{$manifest} = $sub_manifest;

        }

        $RETRY++;

    }
    
    # use the new list of found manifests
    $self->{manifests} = \%found;

}

1;
