package Sophos::ValidatedFetch::Downloader;
use strict;
use URI;
use HTTP::Status qw(RC_OK);
use File::Path qw(mkpath);

=pod
downloader - repo => repo, path => location, [ua => ua]
  getFile($filename) - returns file path
  getContent($filename) - returns content
  repo - url
=cut

our $USERAGENT;
sub _get_ua {
    $USERAGENT ||= do {
      require PPM::PureMessage::UA;
      my $ua = PPM::PureMessage::UA->new;
      return $ua;
    };
    return $USERAGENT;  
}

sub new {
    my ($class,%args) = @_;
    my $repo = delete($args{repo});
    my $path = delete($args{path});
    my $ua = delete($args{ua});
    
    die "path is a required argument" unless $path;
    die "repo is a required argument" unless $repo;

    unless(-d $path) {
        my $status;
        eval { $status = mkpath($path) };
        if(!$status or $@) {
            die "Unable to make directory '$path': $!\n";
        }
    }

    $ua = _get_ua() unless $ua;
    return bless { repo => $repo, 
                   path => $path, 
                   ua => $ua, 
                   error => [],
                 },
                 $class;
  
}


sub _addError {
    my ($self,$error) = @_;
    push(@{$self->{error}},$error);
}

sub error {
    my ($self, %args) = @_;
    my @errors = @{$self->{error}};
    $self->{error} = [];
    return @errors if wantarray;
    return join("\n",@errors);
}

sub getURI {
    my ($self,$uri) = @_;
    $uri = ref($uri) ? $uri : URI->new($uri);
    my @parts = $uri->path_segments;
    my $filename = $parts[-1];
    my $file_location = "$self->{path}/$filename";
    my $ua = $self->{ua};
    my $response = $ua->get($uri, ':content_file' => $file_location);
    if ($response->code() != RC_OK) {
        $self->_addError("Could not fetch '$filename' from repository: " . $response->message());
        return;
    }
    return $file_location; 

}

sub getFile {
    my ($self,$filename) = @_;
 
    my $uri = URI->new($self->{repo});
    $uri->path_segments( $uri->path_segments(), $filename);
    return $self->getURI($uri);
}

sub repo {
    my $self = shift;
    return $self->{repo};
}

1;

