package Sophos::ValidatedFetch::FileCache;
use strict;
use File::Path qw(mkpath);
use ActiveState::File::Atomic;
use File::Basename qw(basename);
use Digest::SHA1;

sub new {
    my ($class,%args) = @_;
    my $path = delete($args{path});
    die "path is required" unless $path;

    $path =~ s{/$}{}; # remove the trailing /

    unless(-d $path) {
        die "Error creating path '$path': $!" unless mkpath($path);
    }

    my %owner = ();
    $owner{owner} = delete($args{uid}) if $args{uid}; 
    $owner{group} = delete($args{gid}) if $args{gid}; 

    die "Unknown option: " . join(',',keys %args) if keys %args;

    return bless { path => $path, ownership => \%owner }, $class;
}

sub _path {
    my ($self,$filename) = @_;
    die "Relative filename is required for $filename" if $filename =~ m{/};
    return $self->{path} . "/" . $filename;
}

sub file_path {
    my ($self,$name) = @_;
    my $file_name = $self->_path($name);
    return unless -e $file_name;
    return $file_name;
}

sub store {
    my ($self,$filepath) = @_;
    return unless -e $filepath;
    my $filename = basename($filepath);
    my $new_path = $self->_path($filename);
    my $atom;

    eval {
      $atom = ActiveState::File::Atomic->new($new_path, 
              writable => 1, 
              create => 1,
              %{$self->{ownership}},
              );
      $atom->commit_file($filepath);
    };

    if ($@) {
        die "Unable to add '$filename' to cache location '$new_path': $@\n";
    }

    unless ( unlink $filepath ) {
        warn "Unable to remove '$filename' after adding to cache located at '$new_path'\n";
    }

    return $new_path;
}

sub getFileContents {
    my ($self,$filename) = @_;
    my ($atom,$contents);
    my $file_path = $self->_path($filename);
    
    return unless -e $file_path;

    eval {
       $atom = ActiveState::File::Atomic->new($file_path); 
       $contents = $atom->slurp();
    }; 
    if($@) {
       die "Unable to read the contents of '$filename': $@\n"; 
    }
    return $contents;
}

sub remove {
    my ($self,$filename) = @_;
    my $file_path = $self->_path($filename);
    return unless -e $file_path;
    unless (unlink $file_path) {
       warn "Unable to unlink '$filename' from our cache at '$file_path': $!";
       return;
    }
    return 1;
}

sub deleteAll {
    my $self = shift;
    my $count = 0;

    if ( opendir(my $dh, $self->{path}) ) {

        while (my $file = readdir($dh) ) {
            my $file_path = $self->_path($file);
            next unless -f $file_path;
            $self->remove($file_path);
            $count++;
        }

        closedir($dh);

    } else {
        die "Unable to access the cache directory '$self->{path}': $!\n";
    }
    $count = '0ButTrue' if $count == 0;
    return $count; 
}

# returns the sha1_hex value
sub sha1 {
    my ($self,$filename) = @_;
    my $file_path = $self->_path($filename);
    return unless -e $file_path;
    my $sh = Digest::SHA1->new();
    my $fv;
    open $fv, $file_path;
    $sh->addfile($fv);
    close $fv;
    return $sh->hexdigest();
}

sub path {
    my $self = shift;
    return $self->{path};
}
=pod
file cache - path => location
  store
  getFileContents
  remove
  deleteAll
  sha1
  path
=cut



1;

