package Sophos::ValidatedFetch::Manifest;
use strict;
=pod

manifest - filename => justName, downloader => $downloader, filecache => $fileCache
  validCertificateChain - boolean if cert contained is correct
  validSignature - boolean if sig is correct
  clear - empty any contents it might have
  getContainedSHA1($name) - returns sha1 of file if it exists, otherwise returns undef
  contents - returns a hash of $name => { sha1 => $sha1, size => $size }
  refresh - retrieves the manifest from the repo, throws exception
  sha1 -???

=cut

sub new {
    my ($class, %args) = @_;

    my $downloader = delete($args{downloader});
    my $filecache  = delete($args{filecache});
    my $filename   = delete($args{filename});
    my $test_no_v  = delete($args{test_no_validate});

    die "downloader is a required argument" unless $downloader;
    die "filecache is a required argument" unless $filecache;
    die "filename is a required argument" unless $filename;


    # data is the raw manifest data
    # contents is the parsed contents of the manifest
    my $self = bless { downloader => $downloader, 
        cache => $filecache, 
        filename => $filename, 
        data => '', 
        contents => {},
        test_no_validate => $test_no_v,
        error => [],
    },$class;

    $self->_load_data();
    return $self;
}

sub _load_data {
    my $self = shift;
    $self->{data} = $self->{cache}->getFileContents( $self->{filename} );
    $self->{data} = '' unless defined($self->{data});
}

sub clear {
    my $self = shift;
    $self->{data} = '';
    $self->{contents} = {};
    $self->{cache}->remove($self->{filename});
    return 1;
}

sub getContainedSHA1 {
    my ($self,$filename) = @_;
    $self->contents() unless scalar( keys(%{$self->{contents}} ) );
    return unless exists($self->{contents}->{$filename});
    return $self->{contents}->{$filename}->{sha1};
}


sub _addError {
    my ($self,$error) = @_;
    push(@{$self->{error}},$error);
}

sub error {
    my ($self, %args) = @_;
    my @errors = @{$self->{error}};
    $self->{error} = [];
    return @errors if wantarray;
    return join("\n",@errors);
}


sub refresh {
    my $self = shift;
    my $newfilename = $self->{downloader}->getFile($self->{filename});
    $self->{contents} = {};
    $self->{data} = undef;
    $self->clear();
    my $error = $self->{downloader}->error();
    $self->_addError($error) if $error;
    return unless $newfilename;
    return unless $self->{cache}->store($newfilename);
    $self->_load_data();
}

sub sha1 {
    my $self = shift;
    return $self->{cache}->sha1($self->{filename});
}

sub contents {
    my $self = shift;
    return () unless $self->{data};
   
    $self->{contents} = {};
    
    my ($manifest,$signature,@certs) = $self->_extract_manifest_elements();
    if ($manifest) {
        # ".\manifest.data" 8241 fa7b4cce52e9efc884883e6e0c27b61e6b87d2ad
        while( $manifest =~ m{^"\.\\([^"]+)"\s(\d+)\s([\da-f]+)}mg ) {
            $self->{contents}->{$1}->{size} = $2;
            $self->{contents}->{$1}->{sha1} = $3;
        }
    }
        
    return %{$self->{contents}};
}

sub _extract_manifest_elements {
    my ($self) = @_;

    my ($manifest, $signature, @certs);

    my $ref = \$manifest;
    for my $line (split /\n/, $self->{data}){
      $line .= "\n";
      if($line =~ m{-----BEGIN SIGNATURE-----}){
        $ref = \$signature;
        next;
      }

      if($line =~ m{-----END SIGNATURE-----}){
        next;
      }

      if($line =~ m{-----BEGIN CERTIFICATE-----}){
        push(@certs, '');
        $ref = \$certs[-1];
      }

      $$ref .= $line;
    }

    return ($manifest, $signature, @certs);
}

sub validSigning {
    my $self = shift;
    if (exists($self->{test_no_validate}) and $self->{test_no_validate}) {
        warn <<EOF;
*************************************************************************
***                NO VALIDATION TEST ENVIRONMENT                     ***
*************************************************************************
EOF
         return 1;
    }
    return unless $self->{data}; 
    
    require Sophos::ValidatedFetch::VerifyManifest;
    my $verifier = Sophos::ValidatedFetch::VerifyManifest->new(manifest => $self->{data});
    eval {
        $verifier->valid();
    };

    if ($@) {
      chomp($@);
      $self->_addError($@);
      return 0;
    }
    return 1;
}


1;


